package ru.dbotthepony.mc.otm.entity

import net.minecraft.core.particles.ParticleTypes
import net.minecraft.server.level.ServerBossEvent
import net.minecraft.server.level.ServerLevel
import net.minecraft.sounds.SoundEvent
import net.minecraft.sounds.SoundEvents
import net.minecraft.sounds.SoundSource
import net.minecraft.util.Mth
import net.minecraft.world.BossEvent
import net.minecraft.world.phys.Vec3
import net.minecraft.world.entity.AnimationState
import net.minecraft.world.entity.EntityType
import net.minecraft.world.entity.LivingEntity
import net.minecraft.world.entity.ai.attributes.AttributeSupplier
import net.minecraft.world.entity.ai.attributes.Attributes
import net.minecraft.world.entity.ai.goal.*
import net.minecraft.world.entity.ai.goal.target.HurtByTargetGoal
import net.minecraft.world.entity.ai.goal.target.NearestAttackableTargetGoal
import net.minecraft.world.entity.ai.navigation.GroundPathNavigation
import net.minecraft.world.entity.ai.navigation.PathNavigation
import net.minecraft.world.entity.monster.*
import net.minecraft.world.entity.npc.Villager
import net.minecraft.world.entity.player.Player
import net.minecraft.world.level.Level

import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.registry.MNames
import ru.dbotthepony.mc.otm.registry.game.MSoundEvents

class Enforcer(type: EntityType<Enforcer>, level: Level) : Monster(type,level) {
	val idleState = AnimationState()
	val chargeState = AnimationState()
	init {
		idleState.start(tickCount)
	}

	override fun registerGoals() {
		goalSelector.addGoal(7, LookAtPlayerGoal(this, Player::class.java, 8f))
		goalSelector.addGoal(3, NearestAttackableTargetGoal(this, LivingEntity::class.java, 10, true, true) { entity ->
			entity is Player ||
				entity is Villager ||
				entity is AbstractIllager ||
				entity is Zombie ||
				entity is AbstractSkeleton
		})

		goalSelector.addGoal(2, RammingGoal(this))
		targetSelector.addGoal(1, HurtByTargetGoal(this))
	}

	override fun createNavigation(level: Level): PathNavigation = GroundPathNavigation(this, level)

	override fun isPushable(): Boolean {
		return false
	}

	override fun getHurtSound(damageSource: net.minecraft.world.damagesource.DamageSource): SoundEvent {
		return SoundEvents.HEAVY_CORE_BREAK
	}

	override fun getDeathSound(): SoundEvent {
		return SoundEvents.VAULT_BREAK
	}
///boss healthbar
	private val bossEvent: ServerBossEvent =
		ServerBossEvent(TranslatableComponent(MNames.ENFORCER), BossEvent.BossBarColor.RED, BossEvent.BossBarOverlay.PROGRESS)

	override fun startSeenByPlayer(player: net.minecraft.server.level.ServerPlayer) {
		super.startSeenByPlayer(player)
		bossEvent.addPlayer(player)
	}

	override fun stopSeenByPlayer(player: net.minecraft.server.level.ServerPlayer) {
		super.stopSeenByPlayer(player)
		bossEvent.removePlayer(player)
	}

	override fun aiStep() {
		super.aiStep()
		bossEvent.progress = this.health / this.maxHealth
	}

	override fun die(cause: net.minecraft.world.damagesource.DamageSource) {
		super.die(cause)
		bossEvent.removeAllPlayers()
	}
///

	//charge attack, could possibly leave a smoke trail too idk
	class RammingGoal(private val mob: Enforcer) : Goal() {
		private var target: LivingEntity? = null

		private var chargeTime = 0
		private val windupTime = 20
		private val maxChargeTime = 40
		private var chargeDir: Vec3? = null
		private var cooldown = 0
		private val minCooldown = 20
		private val maxCooldown = 60
		private var isCharging = false

		override fun canUse(): Boolean {
			if (cooldown > 0) {
				cooldown--
				return false
			}

			target = mob.target
			return target != null
		}

		override fun canContinueToUse(): Boolean {
			return isCharging && chargeTime < maxChargeTime
		}

		override fun start() {
			mob.playSound(MSoundEvents.ENFORCER_ALERT, 1.0f, 1.0f)
			chargeTime = 0
			mob.navigation.stop()
			cooldown = minCooldown + mob.random.nextInt(maxCooldown - minCooldown)
			isCharging = true

			mob.chargeState.start(mob.tickCount)
		}

		override fun tick() {
			if (!isCharging) return

			chargeTime++
			val target = mob.target ?: return

			if (chargeTime < windupTime) {
				chargeDir = Vec3(target.x - mob.x, 0.0, target.z - mob.z).normalize()

				mob.lookControl.setLookAt(target, 30.0f, 30.0f)
				return
			}

			val leftPartic_x = mob.x - mob.lookAngle.x * 2 - 0.8f * mob.lookAngle.z
			val rightPartic_x = mob.x - mob.lookAngle.x * 2 + 0.8f * mob.lookAngle.z
			val smoke_y = mob.y + 0.6f
			val leftPartic_z = mob.z - mob.lookAngle.z * 2 + 0.8f * mob.lookAngle.x
			val rightPartic_z = mob.z - mob.lookAngle.z * 2 - 0.8f * mob.lookAngle.x

			if (mob.level() is ServerLevel) {
				(mob.level() as ServerLevel).sendParticles(
					ParticleTypes.LARGE_SMOKE,
					leftPartic_x, smoke_y, leftPartic_z,
					2,
					0.0, 0.07, 0.0,
					0.01
				)

				(mob.level() as ServerLevel).sendParticles(
					ParticleTypes.LARGE_SMOKE,
					rightPartic_x, smoke_y, rightPartic_z,
					2,
					0.0, 0.07, 0.0,
					0.01
				)
			}

			if (chargeTime == windupTime) {
				if (chargeDir == null || chargeDir!!.length() < 0.1) {
					stop()
					return
				}
				mob.playSound(MSoundEvents.ANDROID_JUMP_BOOST, 1.0f, 1.0f)
			}

			val dir = chargeDir ?: return
			mob.yRot = (-Mth.atan2(dir.x, dir.z) * (180f / Math.PI)).toFloat()
			mob.yHeadRot = mob.yRot

			mob.move(net.minecraft.world.entity.MoverType.SELF, dir.scale(1.8))


			if (mob.horizontalCollision) {
				mob.level().playSound(
					null,
					mob.x, mob.y, mob.z,
					SoundEvents.GENERIC_EXPLODE,
					SoundSource.HOSTILE,
					1.0f,
					1.0f
				)
				stop()
				return
			}

			if (mob.boundingBox.intersects(target.boundingBox)) {
				val knockbackStrength = 1.5
				val knockback = dir.scale(knockbackStrength)

				target.push(knockback.x, 0.5, knockback.z)

				val damageAmount = 6.0f
				target.hurt(mob.damageSources().mobAttack(mob), damageAmount)

				stop()
			}
		}

		override fun stop() {
			mob.setDeltaMovement(Vec3.ZERO)
			chargeTime = 0
			chargeDir = null
			isCharging = false

			mob.chargeState.stop()
		}
	}

	companion object {
		fun createAttributes() : AttributeSupplier.Builder {
			return createMonsterAttributes()
				.add(Attributes.MAX_HEALTH, 300.0)
				.add(Attributes.ARMOR, 20.0)
				.add(Attributes.MOVEMENT_SPEED, 0.3)
				.add(Attributes.STEP_HEIGHT, 1.0)
				.add(Attributes.KNOCKBACK_RESISTANCE, 1.0)
		}
	}
}
